/*
 * $URL$
 * 
 * Copyright (C) 2022 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * $Author$
 *   $Date$
 *    $Rev$
 *
 * (description of this file here)
 *
 */
#include "constants.h"
#include "dark_subtract.h"
#include "config_path.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "special_pixels.h"

void read_coeff_csv(const char* filename, struct shadowcam_image* image, double *coeff_a, double *coeff_b, double *coeff_rmse) {
  FILE *file = fopen(filename, "rb");
  if (file == NULL) {
    printf("fopen of %s failed with errno %d\n", filename, errno);
    exit(1);
  }

  char buffer[1024];

  for (uint8_t channel = 0; channel < SHC_CHANNELS; channel++)
    for (uint16_t column = 0; column < SHC_SCENE; column++) {
      // find a non-comment/header line
      while(fgets(buffer, sizeof(buffer), file))
        if (strncmp(buffer, "#", 1) != 0 && strncmp(buffer, "tdi", 3) != 0)
          break;

      double coeff_tmp[6];

      char *token = strtok(buffer, ",");
      if (token == NULL) {
        printf("Expected ',' in %s, loading CSV failed.\n", filename);
        exit(1);
      }

      for (int i=0; i < 6; i++) {
        coeff_tmp[i] = strtod(token, NULL);
        if (i != 5) {
          token = strtok(NULL, ",");
          if (token == NULL) {
            printf("Expected ',' in %s, loading CSV failed.\n", filename);
            exit(1);
          }
        }
      }

      uint16_t index = channel * SHC_SCENE + column;

      coeff_a[index] = coeff_tmp[0 + 3 * image->tdi_factor];
      coeff_b[index] = coeff_tmp[1 + 3 * image->tdi_factor];
      coeff_rmse[index] = coeff_tmp[2 + 3 * image->tdi_factor];
    }
}

void dark_subtract(struct shadowcam_image* image) {
  puts("Dark subtraction.");

  puts("  Hardcoded to single slope and intercept files! This needs to find the right file once more than one file exists!");

  double *slope_a = malloc(sizeof(double) * SHC_SCENE * SHC_AFE_WIDTH);
  if (slope_a == NULL) {
    printf("malloc failed at %s:%d\n", __FILE__, __LINE__);
    exit(1);
  }
  double *slope_b = malloc(sizeof(double) * SHC_SCENE * SHC_AFE_WIDTH);
  if (slope_b == NULL) {
    printf("malloc failed at %s:%d\n", __FILE__, __LINE__);
    exit(1);
  }
  double *slope_rmse = malloc(sizeof(double) * SHC_SCENE * SHC_AFE_WIDTH);
  if (slope_rmse == NULL) {
    printf("malloc failed at %s:%d\n", __FILE__, __LINE__);
    exit(1);
  }

  puts("  Reading slope coefficients.");
  char *csv_path = config_file_path("dark_subtract_slope.csv");
  read_coeff_csv(csv_path, image, slope_a, slope_b, slope_rmse);
  free(csv_path);

  double *int_a = malloc(sizeof(double) * SHC_SCENE * SHC_AFE_WIDTH);
  if (int_a == NULL) {
    printf("malloc failed at %s:%d\n", __FILE__, __LINE__);
    exit(1);
  }
  double *int_b = malloc(sizeof(double) * SHC_SCENE * SHC_AFE_WIDTH);
  if (int_b == NULL) {
    printf("malloc failed at %s:%d\n", __FILE__, __LINE__);
    exit(1);
  }
  double *int_rmse = malloc(sizeof(double) * SHC_SCENE * SHC_AFE_WIDTH);
  if (int_rmse == NULL) {
    printf("malloc failed at %s:%d\n", __FILE__, __LINE__);
    exit(1);
  }

  puts("  Reading intercept coefficients.");
  csv_path = config_file_path("dark_subtract_intercept.csv");
  read_coeff_csv(csv_path, image, int_a, int_b, int_rmse);
  free(csv_path);

  for (uint32_t line=0; line < image->lines; line++)
    for (uint8_t channel=0; channel < SHC_CHANNELS; channel++)
      for (uint16_t column=0; column < SHC_SCENE; column++) {
        uint16_t sample = column + SHC_SCENE_OFFSET;
        uint16_t ci = channel * SHC_SCENE + column;

        if (!is_special(image->data[data_index(line, channel, sample)])) {
          double dark_level = slope_a[ci] * exp(slope_b[ci] * image->fpa_a_temp) * image->line_rate_ms + int_a[ci] * exp(int_b[ci] * image->fpa_a_temp);

          image->data[data_index(line, channel, sample)] = image->data[data_index(line, channel, sample)] - dark_level;
        }
      }

  free(slope_a);
  free(slope_b);
  free(slope_rmse);
  free(int_a);
  free(int_b);
  free(int_rmse);
}

