/*
 * $URL$
 * 
 * Copyright (C) 2022 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * $Author$
 *   $Date$
 *    $Rev$
 *
 * (description of this file here)
 *
 */
#include <ctype.h>
#include <stdbool.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include "constants.h"
#include "image_struct.h"
#include "read_edr.h"
#include "special_pixels.h"

char *despaceify(char *string) {
  char *end = string + strlen(string) - 1;

  while (isspace((unsigned char)*string))
    string++;

  if (*string == '\0')
    return string;

  while (end > string && isspace((unsigned char)*end))
    end--;
  end[1] = '\0';

  return string;
}

void read_edr_header(struct shadowcam_image *image) {
  puts("Reading EDR header.");

  char buffer[1024];
  bool allgood = false;

  while(fgets(buffer, sizeof(buffer), image->file)) {
    if (strncmp(buffer, "End\n", 4) == 0)
      break;

    if (strncmp(buffer, "Object = IsisCube\n", 18) == 0) {
      allgood = true;
      puts("ISIS cube detected.");
    }

    if (!allgood) {
      printf("%s\n", buffer);
      printf("This thing doesn't look like an ISIS cube.\n");
      exit(1);
    }

    char *keyword = buffer;
    char *value = NULL;

    if ( (value = strchr(buffer, '=')) ) {
      *value = '\0';
      value++;

      keyword = despaceify(keyword);
      value = despaceify(value);
    } else {
      continue;
    }

    if (strcmp(keyword, "StartByte") == 0) {
      image->data_offset = strtol(value, NULL, 10) - 1;
    } else if (strcmp(keyword, "Lines") == 0) {
      image->lines = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "TDIDirection") == 0) {
      image->tdi_direction = value[1];
    } else if (strcmp(keyword, "TemperatureFPAA") == 0) {
      image->fpa_a_temp = strtod(value, NULL);
    } else if (strcmp(keyword, "LineRate") == 0) {
      image->line_rate_ms = strtod(value, NULL);
    } else if (strcmp(keyword, "GainCh0") == 0) {
      image->gain[0] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "OffsetCh0") == 0) {
      image->offset[0] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "GainCh1") == 0) {
      image->gain[1] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "OffsetCh1") == 0) {
      image->offset[1] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "GainCh2") == 0) {
      image->gain[2] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "OffsetCh2") == 0) {
      image->offset[2] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "GainCh3") == 0) {
      image->gain[3] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "OffsetCh3") == 0) {
      image->offset[3] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "GainCh4") == 0) {
      image->gain[4] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "OffsetCh4") == 0) {
      image->offset[4] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "GainCh5") == 0) {
      image->gain[5] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "OffsetCh5") == 0) {
      image->offset[5] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Bterm1") == 0) {
      image->bterm[1] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Bterm2") == 0) {
      image->bterm[2] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Bterm3") == 0) {
      image->bterm[3] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Bterm4") == 0) {
      image->bterm[4] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Bterm5") == 0) {
      image->bterm[5] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Xterm0") == 0) {
      image->xterm[0] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Xterm1") == 0) {
      image->xterm[1] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Xterm2") == 0) {
      image->xterm[2] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Xterm3") == 0) {
      image->xterm[3] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Xterm4") == 0) {
      image->xterm[4] = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "Samples") == 0) {
      image->samples = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "TargetName") == 0) {
      image->target_name = strdup(value);
    } else if (strcmp(keyword, "LineRateCode") == 0) {
      image->line_rate_code = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "PrerollLines") == 0) {
      image->preroll = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "NumberOfKLines") == 0) {
      image->klines = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "ExecutionTime") == 0) {
      image->execution_time = strdup(value);
    } else if (strcmp(keyword, "ExecutionSpacecraftTime") == 0) {
      image->execution_spacecraft_time = strdup(value);
    } else if (strcmp(keyword, "StartTimeOffset") == 0) {
      image->start_time_offset = strtod(value, NULL);
    } else if (strcmp(keyword, "PrerollTime") == 0) {
      image->preroll_time = strdup(value);
    } else if (strcmp(keyword, "StartTime") == 0) {
      image->start_time = strdup(value);
    } else if (strcmp(keyword, "StopTime") == 0) {
      image->stop_time = strdup(value);
    } else if (strcmp(keyword, "TemperatureFPAB") == 0) {
      image->fpa_b_temp = strtod(value, NULL);
    } else if (strcmp(keyword, "TemperatureFPGA") == 0) {
      image->fpga_temp = strtod(value, NULL);
    } else if (strcmp(keyword, "TemperatureTelescope") == 0) {
      image->telescope_temp = strtod(value, NULL);
    } else if (strcmp(keyword, "ModeTest") == 0) {
      image->mode_test = strdup(value);
    } else if (strcmp(keyword, "PDSIdentifier") == 0) {
      image->pds_identifier = strdup(value);
    } else if (strcmp(keyword, "MissionPhase") == 0) {
      image->mission_phase = strdup(value);
    } else if (strcmp(keyword, "DataQualityId") == 0) {
      image->dqi = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "NaifFrameCode") == 0) {
      image->naif_frame_code = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "SampleFirstPixel") == 0) {
      image->sample_first_pixel = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "BiasSpread") == 0) {
      image->bias_spread = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "BiasSpreadCh0") == 0) {
      image->bias_spread_ch0 = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "BiasSpreadCh1") == 0) {
      image->bias_spread_ch1 = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "BiasSpreadCh2") == 0) {
      image->bias_spread_ch2 = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "BiasSpreadCh3") == 0) {
      image->bias_spread_ch3 = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "BiasSpreadCh4") == 0) {
      image->bias_spread_ch4 = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "BiasSpreadCh5") == 0) {
      image->bias_spread_ch5 = strtol(value, NULL, 10);
    } else if (strcmp(keyword, "DQIFpaOutOfBounds") == 0) {
      image->dqi_fpa_out_of_bounds = strdup(value);
    } else if (strcmp(keyword, "DQIUnderExposed") == 0) {
      image->dqi_under_exposed = strdup(value);
    } else if (strcmp(keyword, "DQIMissingData") == 0) {
      image->dqi_missing_data = strdup(value);
    } else if (strcmp(keyword, "DQIMissingSpice") == 0) {
      image->dqi_missing_spice = strdup(value);
    } else if (strcmp(keyword, "DQIUncalibratable") == 0) {
      image->dqi_uncalibratable = strdup(value);
    } else if (strcmp(keyword, "DQICorruptionDetected") == 0) {
      image->dqi_corruption_detected = strdup(value);
    }
  }

  image->tdi_factor = 0;
  if (image->tdi_direction == 'B')
    image->tdi_factor = 1;

  // XXX check if we have what we needed
}

void read_edr_data(struct shadowcam_image *image, bool nospecial) {
  puts("Reading EDR data into 32-bit buffer.");

  uint32_t size = SHC_CHANNELS * SHC_AFE_WIDTH * image->lines;
  uint8_t *buffer = malloc(size);
  if (buffer == NULL) {
    printf("malloc failed at %s:%d\n", __FILE__, __LINE__);
    exit(1);
  }

  if ( (fseek(image->file, image->data_offset, SEEK_SET)) != 0 ) {
    printf("Error seeking to data at offset %u.\n", image->data_offset);
    exit (1);
  }

  size_t lines_read = fread(buffer, SHC_AFE_WIDTH * SHC_CHANNELS, image->lines, image->file);
  if (lines_read != image->lines) {
    printf("Expected %u lines, only read %zu\n", image->lines, lines_read);
    exit (1);
  }

  image->data =  malloc(sizeof(float) * size);
  if (image->data == NULL) {
    printf("malloc failed at %s:%d\n", __FILE__, __LINE__);
    exit(1);
  }

  for(uint32_t i = 0; i < size; i++) {
    if (buffer[i] == 0 && !nospecial)
      image->data[i] = PIXEL_LIS;
    else if (buffer[i] == 255 && !nospecial)
      image->data[i] = PIXEL_HIS;
    else
      image->data[i] = (float) buffer[i];
  }

  free(buffer);
}

void read_edr(struct shadowcam_image *image, bool nospecial) {
  read_edr_header(image);
  ensure_shadowcam_image_populated(image);
  read_edr_data(image, nospecial);
}
